module Main 
where   
open import Foreign.Haskell
open import IO.Primitive
open import Data.String as String using (toCostring)
open import Level using (Level; Lift) renaming 
                                  (zero to 0L; suc to sucL; _⊔_ to _⊔L_)
open import Function using (_$_; _∘_; case_of_; _on_)
open import Function.Equality using (_⟨$⟩_) renaming (_⟶_ to _->c_) 
                                          -- _->c_ is for congruent maps
open import Relation.Nullary.Core using (¬_; Dec; yes; no) 
open import Relation.Binary using 
                        (Rel; Setoid; module Setoid; Decidable; 
                         DecSetoid; module DecSetoid; IsDecEquivalence)
open import Relation.Binary.Core using 
                              (_Preserves₂_⟶_⟶_; _=[_]⇒_; IsEquivalence)
open import Relation.Binary.PropositionalEquality using (_≡_) 
open import Data.Empty using (⊥)
open import Data.Unit  using (⊤)
open import Data.Bool  using (Bool; true; false)
                                  renaming (decSetoid to boolDecSetoid)
open import Data.Maybe    using (Maybe; just; nothing)
open import Data.Product  using (_×_; _,_; ∃)
open import Data.Sum      using (_⊎_)
open import Data.List     using (List; []; _∷_)
open import Data.List.Any using (module Membership) 


------------------------------------------------------------------------
setoidLevel : Level -> Level -> Level
setoidLevel c l =  sucL (c ⊔L l)

data Subset {l} (A : Set l) (member? : A -> Bool) :  Set l
     where
     _cond_ : (a : A) -> (member? a ≡ true) -> Subset A member?

-- Example:  Even = Subset ℕ even?,  (6 cond (even? 6)) : Even.

fromSubset : ∀ {l} {A : Set l} -> {member? : A -> Bool} ->
                                                   Subset A member? -> A
fromSubset (x cond _) = x

boolSetoid : Setoid 0L 0L
boolSetoid = DecSetoid.setoid Data.Bool.decSetoid 

------------------------------------------------------------------------
subsetoid : ∀ {c l} (A : Setoid c l) -> (A ->c boolSetoid) -> Setoid c l

-- The Subsetoid defined by a setoid A and by a (congruent) membership 
--                                                                  map.
subsetoid A member? =  record{ Carrier       = subCarrier
                             ; _≈_           = _=s_
                             ; isEquivalence = isEquiv-s }
      where
      open Setoid A 
      subCarrier = Subset Carrier (\ x -> member? ⟨$⟩ x) 

      _=s_ : Rel subCarrier _
      _=s_ = _≈_ on fromSubset

      isEquiv-s : IsEquivalence _=s_
      isEquiv-s = record{ refl = refl; sym = sym; trans = trans }


subDecSetoid : ∀ {c l} (A : DecSetoid c l) -> 
                    (DecSetoid.setoid A ->c boolSetoid) -> DecSetoid c l

subDecSetoid A member? = record{ Carrier          = sC
                               ; _≈_              = _=s_
                               ; isDecEquivalence = isDecEq-s }
   where
   open DecSetoid A renaming (setoid to baseSetoid)
   sSetoid    = subsetoid baseSetoid member?
   open Setoid sSetoid renaming (Carrier to sC; _≈_ to _=s_; 
                                                isEquivalence to isEq-s)
   _≟s_ : Decidable _=s_
   x ≟s y =  (fromSubset x) ≟ (fromSubset y)

   isDecEq-s : IsDecEquivalence _=s_
   isDecEq-s = record{ isEquivalence = isEq-s;  _≟_ = _≟s_ }


------------------------------------------------------------------------
hasRepetition : ∀ {c l} (A : Setoid c l) -> 
                        List (Setoid.Carrier A) -> Set (setoidLevel c l)
hasRepetition _ []       =  Lift ⊥
hasRepetition A (x ∷ xs) =  x ∈ xs  ⊎ (hasRepetition A xs)
                            where
                            open module MembA = Membership A using (_∈_)

hasFiniteEnumeration : ∀ {c l} -> Setoid c l -> Set _
hasFiniteEnumeration A =  ∃ (\xs -> ¬ hasRepetition A xs) 

DecFiniteEnumeration : ∀ {c l} -> Setoid c l -> Set _
DecFiniteEnumeration =  Dec ∘ hasFiniteEnumeration 

------------------------------------------------------------------------
record DSet c l : Set (setoidLevel c l) 
  where  
  field decSetoid : DecSetoid c l

  setoid : Setoid c l
  setoid = DecSetoid.setoid decSetoid
     
  Carrier : Set c
  Carrier = Setoid.Carrier setoid

  field  mbFiniteEnum : Maybe $ DecFiniteEnumeration setoid

------------------------------------------------------------------------
subDSet : ∀ {c l} (dSet : DSet c l) -> 
                 (member? : DSet.setoid dSet ->c boolSetoid) -> DSet c l 
-- Sub-d-set  defined by dSet and a membership predicate.

subDSet dSet member? =  record{ decSetoid    = sDecSetoid
                              ; mbFiniteEnum = nothing }
   where
   sDecSetoid = subDecSetoid (DSet.decSetoid dSet) member?

------------------------------------------------------------------------
record SubDSet c l :  Set (setoidLevel c l) 
  where
  -- SubDSet  is a base DSet,  membership predicate,
  -- sub-DSet = subDSet baseDSet member?,   and certain more attributes.

  field  baseDSet : DSet c l   --

  baseDecSetoid = DSet.decSetoid baseDSet
  baseSetoid    = DecSetoid.setoid baseDecSetoid
  open Setoid baseSetoid using (_≈_) renaming (Carrier to baseCarrier)

  field  member? : baseSetoid ->c boolSetoid      -- a congruent map

  sub-DSet : DSet c l
  sub-DSet = subDSet baseDSet member?

  sub-DecSetoid : DecSetoid c l
  sub-DecSetoid = DSet.decSetoid sub-DSet

  sub-Setoid : Setoid c l
  sub-Setoid = DecSetoid.setoid sub-DecSetoid
  open Setoid sub-Setoid renaming (Carrier to subCarrier; _≈_ to _=s_)

  memberC? : baseCarrier -> Bool
  memberC? x =  member? ⟨$⟩ x     

  fromSub : subCarrier -> baseCarrier
  fromSub = fromSubset {c} {baseCarrier} {memberC?} 

  eTo : sub-Setoid ->c baseSetoid 
                     -- Π sub-Setoid (B.Setoid.indexedSetoid baseSetoid)
  eTo = record{ _⟨$⟩_ = fromSub 
              ; cong  = eCong   }  -- <-- "unsolved metas"
        where
        postulate eCong :  _=s_ =[ fromSub ]⇒ _≈_

  -- why  "cong = eCong"  is said  "unsolved metas" ?

        {- this variant is type checked, but does not help with the 
           above "unsolved metas" :

        eCong {x} {y} x-=s-y =  x'=y'      -- P =[ f ]⇒ Q = P ⇒ (Q on f) 
                                where
                                x' = fromSub x
                                y' = fromSub y
                                x'=y' : x' ≈ y'
                                x'=y' = x-=s-y
        -}

------------------------------------------------------------------------
main : IO Unit
main = putStrLn $ toCostring "done"
